#include <stdio.h>
#include <stdlib.h>
#include "util.h"
#include "bitstream.h"
#include "guessfsb.h"
#include "fsbext.h"
#include "riffext.h"
#include "xma_rebuild.h"

// XMAsh - decrypt, demux, and rebuild FSB XMA2
// also minimal RIFF support
// cobbled together from:
// romchu, guessfsb, decfsb, fsb_mpeg, xma_parse, vgmstream

#define VERSION "0.1"
#define BIN_NAME "xmash"

struct main_info {
    const char *file_name;
    long block_size; // -1 for nonblocked
};

int try_key_callback(const uint8_t * d, long size, void *v);
int subfile_callback(const uint8_t * infile, long size, int streams, long samples, long srate, const char *stream_name, void *v);

void usage(void);

int main(int argc, char **argv)
{
    long file_size;
    uint8_t *indata;

    int success = 0;

    struct main_info mi;

    if (argc < 2)
    {
        usage();
    }

    char * infile_name = argv[1];

    mi.file_name = strip_path(infile_name);
    mi.block_size = default_block_size;

    for (int argno = 2; argno < argc; argno++) {
        if (argv[argno][0] == '-' &&
            argv[argno][1] != '\0' &&
            argv[argno][2] == '\0')
        {
            switch (argv[argno][1]) {
                case 'b':
                    CHECK_ERROR(argno + 1 >= argc, "-b needs an argument");

                    mi.block_size = read_long(argv[++argno]);
                    CHECK_ERROR(mi.block_size <= 0, "invalid block size");

                    break;
                case 'B':
                    // use data size (nonblocked)
                    mi.block_size = -1;
                    break;
                default:
                    usage();
                    break;
            }
        }
        else
        {
            usage();
        }
    }

    {
        FILE *infile;
        infile = fopen(infile_name, "rb");
        CHECK_ERRNO(!infile, "open input failed");
        indata = get_whole_file(infile, &file_size);
        fclose(infile);
    }

    printf("%s\n\n", infile_name);

    // try without decrypting
    if (0 == try_multistream_fsb(indata, file_size, subfile_callback, &mi))
    {
        printf("success without decryption!\n");
        success = 1;
    }
    // try RIFF
    else if (0 == try_xma_riff(indata, file_size, subfile_callback, &mi))
    {
        printf("success with RIFF!\n");
        success = 1;
    }
    // try via guessfsb
    else if (0 == guess_fsb_keys(indata, file_size, try_key_callback, &mi))
    {
        printf("success!\n");
        success = 1;
    }
    else
    {
        printf("failure.\n");
    }

    free(indata);

    if (success)
    {
        return 0;
    }
    else
    {
        return 1;
    }
}

void usage()
{
    fprintf(stderr, "XMAsh " VERSION " - decrypt, demux, and rebuild FSB (or RIFF) XMA2\n");
    fprintf(stderr, "usage:\n"
                    "  " BIN_NAME " input.fsb.xen [options]\n"
                    "  " BIN_NAME " input.fsb [options]\n"
                    "  " BIN_NAME " input.xma [options]\n"
                    "options:\n"
                    "  -b XMA2 block size (default 0x%lx)\n"
                    "  -B XMA2 is nonblocked\n",
                    (unsigned long)default_block_size);
    exit(EXIT_FAILURE);
}

// called for each possibly valid key
int try_key_callback(const uint8_t * indata, long size, void *v)
{
    return try_multistream_fsb(indata, size, subfile_callback, v);
}

// called for each subfile in an FSB (or a RIFF body)
int subfile_callback(const uint8_t * infile, long size, int streams, long samples, long srate, const char *stream_name, void *v)
{
    const struct main_info *mip = v;

    // build the name
    int name_base_len = strlen(mip->file_name) + 1 + strlen(stream_name);
    char *name_base = malloc(name_base_len+1);
    CHECK_ERRNO(!name_base, "malloc");
    sprintf(name_base, "%s_%s", mip->file_name, stream_name);

    // rebuild each stream
    for (int str = 0; str < streams; str ++)
    {
        long parsed_samples;
        long data_offset, data_size, block_size;
        char *strname = number_name(name_base, ".xma", str+1, streams);
        printf("%s\n", strname);

        FILE *outfile = fopen(strname, "wb");
        CHECK_ERRNO(!outfile, "fopen output");
        CHECK_ERRNO(
            -1 == fseek(outfile, xma_header_size, SEEK_SET), "fseek past header");

        data_offset = packet_size_bytes*str;
        data_size = size-data_offset;
        block_size = mip->block_size;
        // nonblocked, use whole file as block size
        if (-1 == mip->block_size)
        {
            block_size = data_size;
        }

        // rebuild!
        if (0 != build_XMA_from_XMA2(infile + data_offset,
                                     data_size,
                                     outfile, block_size, 2,
                                     &parsed_samples))
        {
            // encountered an error while parsing
            fclose(outfile);
            free(strname);
            free(name_base);
            return 1;
        }

        parsed_samples += 128; // seems to always be short by 128

        if (-1 != samples && parsed_samples != samples)
        {
            fprintf(stderr,"parsed samples = %ld, expected %ld\n", parsed_samples, samples);
            fclose(outfile);
            free(strname);
            free(name_base);
            return 1;
        }

        // write header
        long finish_offset = ftell(outfile);
        uint8_t *xma_head = make_xma_header(srate, finish_offset-xma_header_size, 2);
        CHECK_ERRNO(
            -1 == fseek(outfile, 0, SEEK_SET), "fseek to header");
        CHECK_ERRNO( 1 != fwrite(xma_head, xma_header_size, 1, outfile) , "fwrite header");

        CHECK_ERRNO(EOF == fclose(outfile), "fclose");
        free(xma_head);

        free(strname);
    }

    free(name_base);
    return 0;
}


